package Bio::MUST::Apps::Physeter;
# ABSTRACT: Dummy internal class for physeter tool
$Bio::MUST::Apps::Physeter::VERSION = '0.213470';
use strict;
use warnings;

1;

__END__

=pod

=head1 NAME

Bio::MUST::Apps::Physeter - Dummy internal class for physeter tool

=head1 VERSION

version 0.213470

=head1 AUTHOR

Denis BAURAIN <denis.baurain@uliege.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2020 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
